@@ -5,10 +5,14 @@ import io
 import contextlib
 import functools
 import logging
+import warnings
 
 import boto3
 import botocore.client
 import six
+import sys
+
+import smart_open.bytebuffer
 
 logger = logging.getLogger(__name__)
 
@@ -20,7 +24,7 @@ try:
     import multiprocessing.pool
     _MULTIPROCESSING = True
 except ImportError:
-    logger.warning("multiprocessing could not be imported and won't be used")
+    warnings.warn("multiprocessing could not be imported and won't be used")
 
 
 DEFAULT_MIN_PART_SIZE = 50 * 1024**2
@@ -32,6 +36,11 @@ WRITE_BINARY = 'wb'
 MODES = (READ_BINARY, WRITE_BINARY)
 """Allowed I/O modes for working with S3."""
 
+_BINARY_TYPES = (six.binary_type, bytearray)
+"""Allowed binary buffer types for writing to the underlying S3 stream"""
+if sys.version_info >= (2, 7):
+    _BINARY_TYPES = (six.binary_type, bytearray, memoryview)
+
 BINARY_NEWLINE = b'\n'
 
 SUPPORTED_SCHEMES = ("s3", "s3n", 's3u', "s3a")
@@ -64,8 +73,8 @@ def open(
         buffer_size=DEFAULT_BUFFER_SIZE,
         min_part_size=DEFAULT_MIN_PART_SIZE,
         session=None,
-        resource_kwargs=dict(),
-        multipart_upload_kwargs=dict(),
+        resource_kwargs=None,
+        multipart_upload_kwargs=None,
         ):
     """Open an S3 object for reading or writing.
 
@@ -84,7 +93,7 @@ def open(
     session: object, optional
         The S3 session to use when working with boto3.
     resource_kwargs: dict, optional
-        Keyword arguments to use when creating a new resource.  For writing only.
+        Keyword arguments to use when accessing the S3 resource for reading or writing.
     multipart_upload_kwargs: dict, optional
         Additional parameters to pass to boto3's initiate_multipart_upload function.
         For writing only.
@@ -94,6 +103,11 @@ def open(
     if mode not in MODES:
         raise NotImplementedError('bad mode: %r expected one of %r' % (mode, MODES))
 
+    if resource_kwargs is None:
+        resource_kwargs = {}
+    if multipart_upload_kwargs is None:
+        multipart_upload_kwargs = {}
+
     if mode == READ_BINARY:
         fileobj = SeekableBufferedInputBase(
             bucket_id,
@@ -156,7 +170,7 @@ class SeekableRawReader(object):
         #
         try:
             self._body.close()
-        except AttributeError as e:
+        except AttributeError:
             pass
 
         if position == self._content_length == 0 or position == self._content_length:
@@ -181,17 +195,19 @@ class SeekableRawReader(object):
 
 class BufferedInputBase(io.BufferedIOBase):
     def __init__(self, bucket, key, buffer_size=DEFAULT_BUFFER_SIZE,
-                 line_terminator=BINARY_NEWLINE, session=None, resource_kwargs=dict()):
+                 line_terminator=BINARY_NEWLINE, session=None, resource_kwargs=None):
         if session is None:
             session = boto3.Session()
+        if resource_kwargs is None:
+            resource_kwargs = {}
+
         s3 = session.resource('s3', **resource_kwargs)
         self._object = s3.Object(bucket, key)
         self._raw_reader = RawReader(self._object)
         self._content_length = self._object.content_length
         self._current_pos = 0
-        self._buffer = b''
+        self._buffer = smart_open.bytebuffer.ByteBuffer(buffer_size)
         self._eof = False
-        self._buffer_size = buffer_size
         self._line_terminator = line_terminator
 
         #
@@ -226,10 +242,7 @@ class BufferedInputBase(io.BufferedIOBase):
         if size == 0:
             return b''
         elif size < 0:
-            if len(self._buffer):
-                from_buf = self._read_from_buffer(len(self._buffer))
-            else:
-                from_buf = b''
+            from_buf = self._read_from_buffer()
             self._current_pos = self._content_length
             return from_buf + self._raw_reader.read()
 
@@ -243,7 +256,7 @@ class BufferedInputBase(io.BufferedIOBase):
         # If the stream is finished, return what we have.
         #
         if self._eof:
-            return self._read_from_buffer(len(self._buffer))
+            return self._read_from_buffer()
 
         #
         # Fill our buffer to the required size.
@@ -272,19 +285,20 @@ class BufferedInputBase(io.BufferedIOBase):
         the_line = io.BytesIO()
         while not (self._eof and len(self._buffer) == 0):
             #
-            # In the worst case, we're reading self._buffer twice here, once in
-            # the if condition, and once when calling index.
+            # In the worst case, we're reading the unread part of self._buffer
+            # twice here, once in the if condition and once when calling index.
             #
             # This is sub-optimal, but better than the alternative: wrapping
             # .index in a try..except, because that is slower.
             #
-            if self._line_terminator in self._buffer:
-                next_newline = self._buffer.index(self._line_terminator)
+            remaining_buffer = self._buffer.peek()
+            if self._line_terminator in remaining_buffer:
+                next_newline = remaining_buffer.index(self._line_terminator)
                 the_line.write(self._read_from_buffer(next_newline + 1))
                 break
             else:
-                the_line.write(self._read_from_buffer(len(self._buffer)))
-                self._fill_buffer(self._buffer_size)
+                the_line.write(self._read_from_buffer())
+                self._fill_buffer()
         return the_line.getvalue()
 
     def terminate(self):
@@ -294,22 +308,20 @@ class BufferedInputBase(io.BufferedIOBase):
     #
     # Internal methods.
     #
-    def _read_from_buffer(self, size):
+    def _read_from_buffer(self, size=-1):
         """Remove at most size bytes from our buffer and return them."""
         # logger.debug('reading %r bytes from %r byte-long buffer', size, len(self._buffer))
-        assert size >= 0
-        part = self._buffer[:size]
-        self._buffer = self._buffer[size:]
+        size = size if size >= 0 else len(self._buffer)
+        part = self._buffer.read(size)
         self._current_pos += len(part)
         # logger.debug('part: %r', part)
         return part
 
-    def _fill_buffer(self, size):
+    def _fill_buffer(self, size=-1):
+        size = size if size >= 0 else self._buffer._chunk_size
         while len(self._buffer) < size and not self._eof:
-            raw = self._raw_reader.read(size=self._buffer_size)
-            if len(raw):
-                self._buffer += raw
-            else:
+            bytes_read = self._buffer.fill(self._raw_reader)
+            if bytes_read == 0:
                 logger.debug('reached EOF while filling buffer')
                 self._eof = True
 
@@ -320,17 +332,18 @@ class SeekableBufferedInputBase(BufferedInputBase):
     Implements the io.BufferedIOBase interface of the standard library."""
 
     def __init__(self, bucket, key, buffer_size=DEFAULT_BUFFER_SIZE,
-                 line_terminator=BINARY_NEWLINE, session=None, resource_kwargs=dict()):
+                 line_terminator=BINARY_NEWLINE, session=None, resource_kwargs=None):
         if session is None:
             session = boto3.Session()
+        if resource_kwargs is None:
+            resource_kwargs = {}
         s3 = session.resource('s3', **resource_kwargs)
         self._object = s3.Object(bucket, key)
         self._raw_reader = SeekableRawReader(self._object)
         self._content_length = self._object.content_length
         self._current_pos = 0
-        self._buffer = b''
+        self._buffer = smart_open.bytebuffer.ByteBuffer(buffer_size)
         self._eof = False
-        self._buffer_size = buffer_size
         self._line_terminator = line_terminator
 
         #
@@ -366,7 +379,7 @@ class SeekableBufferedInputBase(BufferedInputBase):
         self._raw_reader.seek(new_position)
         logger.debug('new_position: %r', self._current_pos)
 
-        self._buffer = b""
+        self._buffer.empty()
         self._eof = self._current_pos == self._content_length
         return self._current_pos
 
@@ -389,10 +402,9 @@ class BufferedOutputBase(io.BufferedIOBase):
             bucket,
             key,
             min_part_size=DEFAULT_MIN_PART_SIZE,
-            s3_upload=None,
             session=None,
-            resource_kwargs=dict(),
-            multipart_upload_kwargs=dict(),
+            resource_kwargs=None,
+            multipart_upload_kwargs=None,
             ):
         if min_part_size < MIN_MIN_PART_SIZE:
             logger.warning("S3 requires minimum part size >= 5MB; \
@@ -400,6 +412,10 @@ multipart upload may fail")
 
         if session is None:
             session = boto3.Session()
+        if resource_kwargs is None:
+            resource_kwargs = {}
+        if multipart_upload_kwargs is None:
+            multipart_upload_kwargs = {}
 
         s3 = session.resource('s3', **resource_kwargs)
 
@@ -477,8 +493,10 @@ multipart upload may fail")
 
         There's buffering happening under the covers, so this may not actually
         do any HTTP transfer right away."""
-        if not isinstance(b, six.binary_type):
-            raise TypeError("input must be a binary string, got: %r", b)
+
+        if not isinstance(b, _BINARY_TYPES):
+            raise TypeError(
+                "input must be one of %r, got: %r" % (_BINARY_TYPES, type(b)))
 
         self._buf.write(b)
         self._total_bytes += len(b)
